from base64 import b64encode, b64decode, urlsafe_b64encode
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.kdf.scrypt import Scrypt
import os
import random
import string

key_size = 4096
msg_size = 446
salt_size = 16


class Pto:
    "cry.Pto see what I did there?"

    def __init__(self, pem, password: str = None):
        if password:
            # pem is private key
            password = password.encode()
            pem = b64decode(pem.encode())
            salt = pem[:salt_size]
            encrypted_pvt = pem[salt_size:]
            fernet_key = Scrypt(salt=salt, length=32, n=2**18, r=8, p=1).derive(
                password
            )
            fernet = Fernet(urlsafe_b64encode(fernet_key))
            pvt = fernet.decrypt(encrypted_pvt)
            self.private_key = serialization.load_pem_private_key(
                data=pvt,
                backend=default_backend(),
                password=None,
            )
            self.public_key = self.private_key.public_key()
            return
        # pem is public key
        self.public_key = serialization.load_pem_public_key(
            data=pem.encode(), backend=default_backend()
        )

    def encrypt(self, message: str):
        bmsg = message.encode()
        ciphertext = bytes()
        for chunk in [bmsg[i : i + msg_size] for i in range(0, len(bmsg), msg_size)]:
            ciphertext += self.public_key.encrypt(
                chunk,
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None,
                ),
            )
        return b64encode(ciphertext).decode("utf-8")

    def decrypt(self, ciphertext: str):
        bciph = b64decode(ciphertext.encode())
        message = bytes()
        for chunk in [
            bciph[i : i + key_size // 8] for i in range(0, len(bciph), key_size // 8)
        ]:
            message += self.private_key.decrypt(
                chunk,
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None,
                ),
            )
        return message.decode("utf-8")


def generate_key_pair(password: str):
    password = password.encode()
    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=key_size, backend=default_backend()
    )
    public_key = private_key.public_key()
    pem_pvt = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    pem_pub = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    salt = os.urandom(salt_size)
    fernet_key = Scrypt(salt=salt, length=32, n=2**18, r=8, p=1).derive(password)
    fernet = Fernet(urlsafe_b64encode(fernet_key))
    encrypted_pvt = fernet.encrypt(pem_pvt)
    serialized_pvt = b64encode(salt + encrypted_pvt).decode("utf-8")
    serialized_pub = pem_pub.decode("utf-8")
    return (serialized_pvt, serialized_pub)


def make_password(length=16):
    selectable_punctuation = "+-=_&?/\\%#@~,."
    chars = (
        string.ascii_lowercase
        + string.ascii_uppercase
        + string.digits
        + selectable_punctuation
    )
    password = "".join(random.choice(chars) for x in range(length))
    return password



"""
loz functions
"""
from loz import cry
from loz.exceptions import *
from os import path
import csv
import io
import json
import logging
import secrets

logger = logging.getLogger("loz")
storage_version = "1.0"

global pto


def load(loz_file):
    "Loads .loz file and returns storage object."
    global pto
    if not path.isfile(loz_file):
        raise LozFileDoesNotExist(f"No loz file found at {loz_file}.")
    storage = None
    logger.debug(f"opening {loz_file}")
    with open(loz_file, "r") as f:
        storage = json.loads(f.read())
        logger.debug(f"loaded storage from {loz_file}")
        if storage["version"] > storage_version:
            raise LozIncompatibleFileVersion(
                f"Found lozfile version {storage['version']} but supports up to {storage_version}."
            )
    pto = cry.Pto(storage["pub"])
    return storage


def save(storage, loz_file):
    "Saves storage object as json .loz file."
    logger.debug(f"opening {loz_file}")
    storage["version"] = storage_version
    with open(loz_file, "w") as f:
        f.write(json.dumps(storage))
        logger.debug(f"storage written to {loz_file}")


def validate_key(storage, password=None):
    "Tries to unlock storage with a password. Store key as global for later use."
    global pto
    try:
        logger.debug("validating password")
        pto = cry.Pto(storage["pvt"], password)
        return True
    except Exception as e:
        print(e)
        return False


def init(password):
    "Initialize storage with master password."
    pvt, pub = cry.generate_key_pair(password)
    storage = {"pvt": pvt, "pub": pub, "data": {}}
    return storage


def export_csv(storage, password, csv_file=None):
    "Export plaintext storage and print to output or save to CSV file."
    return_contents = False
    if not csv_file:
        csv_file = io.StringIO()
        return_contents = True
    writer = csv.writer(
        csv_file, delimiter=" ", quotechar="|", quoting=csv.QUOTE_MINIMAL
    )
    for row in get_all(storage, password):
        writer.writerow(row)
    if return_contents:
        return csv_file.getvalue()


def import_csv(storage, csv_file):
    rows = csv.reader(csv_file, delimiter=" ", quotechar="|")
    total_change = ["", "", "", "total changes"]
    for row in rows:
        domain = row[0]
        user = row[1]
        secret = row[2]
        change = store(storage, domain, user, secret)
        yield change


def change_password(storage, old_pass, new_pass):
    pvt, pub = cry.generate_key_pair(new_pass)
    new_storage = {"pvt": pvt, "pub": pub, "data": {}}
    new_pto = cry.Pto(new_storage["pvt"], new_pass)
    for domain, user, secret in get_all(storage, old_pass):
        logger.debug(f"moving {domain}/{user} to new storage")
        encrypted = new_pto.encrypt(secret)
        if domain in new_storage["data"].keys():
            new_storage["data"][domain][user] = encrypted
        else:
            new_storage["data"][domain] = {user: encrypted}
    return new_storage


def store(storage, domain, user, secret=None):
    "Store secret for domain/user pair. Return [change_type, domain, user, message] list of changed item."
    global pto
    if not secret:
        secret = cry.make_password()
        print(secret)
    encrypted = pto.encrypt(secret)
    if exists(storage, domain, user):  # update existing
        logger.debug(f"updating secret for {domain}/{user}")
        storage["data"][domain][user] = encrypted
        return ["-+", domain, user, "Update existing secret."]
    if exists(storage, domain):  # add new user
        logger.debug(f"adding user {user} with secret for {domain}")
        storage["data"][domain][user] = encrypted
        return ["+", domain, user, "Add user and secret."]
    # add user and domain
    logger.debug(f"adding domain {domain} and user {user} with secret")
    storage["data"][domain] = {user: encrypted}
    return ["+", domain, user, "Add domain, user and secret."]


def exists(storage, domain, user=None):
    if domain not in storage["data"]:
        logger.debug(f"domain {domain} does not exist")
        return False
    if not user:
        logger.debug(f"domain {domain} exists")
        return True
    if user in storage["data"][domain]:
        logger.debug(f"user {user} exists in domain {domain}")
        return True
    logger.debug(f"user {user} does not exist in domain {domain}")
    return False


def get(storage, domain, user):
    "Return secret for 'domain user' pair."
    global pto
    if not exists(storage, domain, user):
        logger.warning(f"{domain}/{user} does not exist")
        return None
    secret = storage["data"][domain][user]
    return pto.decrypt(secret)


def get_all(storage, password):
    for domain in storage["data"].keys():
        for user in storage["data"][domain].keys():
            yield [
                domain,
                user,
                get(storage, domain, user),
            ]


def rm(storage, domain, user=None):
    "Delete user from domain or whole domain, return change."
    if not exists(storage, domain):
        return ["|", "", "", "No change."]
    if user and not exists(storage, domain, user):
        return ["|", "", "", "No change."]
    if user:
        del storage["data"][domain][user]
        if not storage["data"][domain]:
            del storage["data"][domain]
            return ["-", domain, "*", "Delete whole domain."]
        return ["-", domain, user, "Delete user."]
    else:
        del storage["data"][domain]
        return ["-", domain, "*", "Delete whole domain."]


def show(storage, domain, password):
    "List all secrets in one domain."
    global pto
    if domain not in storage["data"]:
        return None
    for user in storage["data"][domain]:
        logger.debug("user found, decrypting secret")
        secret = storage["data"][domain][user]
        decrypted = pto.decrypt(secret)
        yield [domain, user, decrypted]


def ls(storage, domain=None):
    "List all storage or users under one domain."
    if not domain:
        return list(storage["data"].keys())
    if domain not in storage["data"]:
        return []
    return list(storage["data"][domain].keys())


def find(storage, word):
    "List storage and users that contain the search phrase."
    results = []
    for domain in storage["data"]:
        if word in domain:
            results.append(
                [
                    domain,
                ]
            )
        for user in storage["data"][domain]:
            if word in user:
                results.append([domain, user])
    return results



"""
loz main module.
"""
from click_aliases import ClickAliasedGroup
from getpass import getpass
from loz import loz
from loz.exceptions import *
from os import path
from prompt_toolkit import prompt
import click
import csv
import logging
import loz as app
import sys

logging.basicConfig()
logger = logging.getLogger("loz")
logger.setLevel(logging.WARNING)

loz_file = path.expanduser("~/.loz")
check_mode = False
yes_mode = False


class color:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


def load(warn=True):
    storage = None
    try:
        storage = loz.load(loz_file)
    except LozFileDoesNotExist as e:
        logger.debug(e)
        if warn:
            print(f"First time use. Run init command.")
        sys.exit(1)
    except LozIncompatibleFileVersion as e:
        logger.debug(e)
        if warn:
            print(f"Update loz. {e}")
        sys.exit(1)
    return storage


def confirm_change(change):
    if check_mode or yes_mode:
        return True
    if not "-" in change[0]:
        # non-destructive change
        return True
    answer = input("Is this ok [Y/n]: ")
    if answer.lower() in ["", "y", "yes"]:
        print("Confirmed!")
        return True
    print("Abandoning changes.")
    return False


def print_change(change):
    col = color.GREEN
    if "+" in change[0]:
        col = color.YELLOW
    if "-" in change[0]:
        col = color.RED
    print(
        f"{col}{change[0]} {color.BOLD}{change[1]}/{change[2]}{color.END} {col}{change[3]}{color.END}"
    )


def save(storage, change):
    print_change(change)
    if not check_mode:
        if confirm_change(change):
            loz.save(storage, loz_file)
        return
    print("Check mode is on, so nothing is really changed.")


def enter_password(storage):
    while True:
        password = getpass("enter password:")
        if loz.validate_key(storage, password):
            return password


def enter_new_password():
    password = None
    while password == None:
        password = getpass("enter new password:")
        if password != getpass("repeat:"):
            print(f"passwords do not match")
            password = None
    return password


@click.group(cls=ClickAliasedGroup)
@click.version_option(
    version=app.__version__, message=f"%(prog)s %(version)s - {app.__copyright__}"
)
@click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Enable debug mode with output of each action in the log.",
)
@click.option(
    "-c",
    "--check",
    is_flag=True,
    help="Enable check mode that shows changes but changes nothing.",
)
@click.option(
    "-y",
    "--yes",
    is_flag=True,
    help="Answer yes to all questions.",
)
@click.option(
    "-f",
    "--file",
    type=str,
    default=loz_file,
    help=f"Specify custom loz file. Defaults to: {loz_file}",
)
@click.pass_context
def cli(ctx, **kwargs):
    global loz_file
    global check_mode
    global yes_mode
    if ctx.params.get("debug"):
        logger.setLevel(logging.DEBUG)
        logger.info("debug mode is on")
    check_mode = ctx.params.get("check")
    yes_mode = ctx.params.get("yes")
    loz_file = ctx.params.get("file")


@cli.command()
def init():
    "Initialize .loz file with master password."
    change = ["", "", "", "Initialize lozfile with password."]
    if path.isfile(loz_file):
        change = ["+-", "", "", "Overwrite lozfile and reinitialize."]
    password = enter_new_password()
    storage = loz.init(password)
    save(storage, change)


@cli.command()
def passwd():
    "Change password and re-encrypt storage."
    storage = load(loz_file)
    password = enter_password(storage)
    new_password = enter_new_password()
    new_storage = loz.change_password(storage, password, new_password)
    loz.save(new_storage, loz_file)


@cli.command(aliases=["set"])
@click.argument("domain")
@click.argument("user")
def add(domain, user):
    "Set secret for 'domain user' pair."
    storage = load(loz_file)
    secret = prompt(
        "Write multiline secret. Alt+Return to save:\n", default="", multiline=True
    )
    change = loz.store(storage, domain, user, secret)
    save(storage, change)


@cli.command()
@click.argument("domain")
@click.argument("user")
def make(domain, user):
    "Make random secret for 'domain user' pair."
    storage = load(loz_file)
    change = loz.store(storage, domain, user)
    save(storage, change)


@cli.command()
@click.argument("domain")
@click.argument("user")
def get(domain, user):
    "Get secret for 'domain user' pair."
    storage = load(loz_file)
    if not loz.exists(storage, domain, user):
        print(f"user not found: {color.BOLD}{domain}/{user}{color.END}")
        sys.exit(1)
    password = enter_password(storage)
    secret = loz.get(storage, domain, user)
    print(secret)


@cli.command(aliases=["del"])
@click.argument("domain")
@click.argument("user", required=False, default=None)
def rm(domain, user):
    "Delete user from domain or whole domain."
    storage = load(loz_file)
    change = loz.rm(storage, domain, user)
    save(storage, change)


@cli.command()
@click.argument("domain")
def show(domain):
    "List all secrets in one domain."
    storage = load(loz_file)
    if not loz.exists(storage, domain):
        print(f"domain not found: {color.BOLD}{domain}{color.END}")
        sys.exit(1)
    password = enter_password(storage)
    for domain, user, secret in loz.show(storage, domain, password):
        print(f"\n{color.BOLD}{domain}/{user}{color.END}\n{secret}")


@cli.command()
@click.argument("domain", required=False, default=None)
def ls(domain):
    "List all storage or users under one domain."
    storage = load(loz_file)
    results = loz.ls(storage, domain)
    print("\n".join(results))


@cli.command(aliases=["search"])
@click.argument("word")
def find(word):
    "List storage and users that contain the search phrase."
    storage = load(loz_file)
    results = loz.find(storage, word)
    for res in results:
        bold = f"{color.BOLD}{res[-1]}{color.END}"
        res[-1] = bold
        print("/".join(res))


@cli.command("import")
@click.argument("csv_file", type=click.File("r"))
def import_csv(csv_file):
    "Import CSV file and merge with existing lozfile."
    storage = load(loz_file)
    total_change = ["", "", "", "total change"]
    for change in loz.import_csv(storage, csv_file):
        print_change(change)
        total_change[0] += change[0]
    if not total_change[0]:
        print("\bNo changes detected.")
        return
    save(storage, total_change)


@cli.command("export")
@click.argument("csv_file", type=click.File("w"), required=False, default=sys.stdout)
def export_csv(csv_file):
    "Export plaintext storage and print to output or save to CSV file."
    storage = load(loz_file)
    password = enter_password(storage)
    loz.export_csv(storage, password, csv_file)


@cli.command()
def bash_completion():
    "Generate bash completion file contents."
    completion = (
        f"_loz()\n"
        f"{{\n"
        f"    local cur prev\n"
        f"    COMPREPLY=()\n"
        f'    cur="${{COMP_WORDS[COMP_CWORD]}}"\n'
        f'    prev="${{COMP_WORDS[COMP_CWORD - 1]}}"\n'
        f"    words=\n"
        f"\n"
        f'    if [ "$COMP_CWORD" == "1" ]; then\n'
        f'        words="{" ".join([n for n, v in cli.commands.items()])}"\n'
        f"    fi\n"
        f"\n"
        f'    cmd="${{COMP_WORDS[1]}}"\n'
        f'    case "$cmd" in\n'
        f"        get|set|add|make|rm|del)\n"
        f'            case "$COMP_CWORD" in\n'
        f"                2)\n"
        f'                    words="$(loz ls)"\n'
        f"                    ;;\n"
        f"                3)\n"
        f'                    words="$(loz ls $prev)"\n'
        f"                    ;;\n"
        f"            esac\n"
        f"            ;;\n"
        f"        show|ls)\n"
        f'            if [ "$COMP_CWORD" == "2" ]; then\n'
        f'                words="$(loz ls)"\n'
        f"            fi\n"
        f"            ;;\n"
        f"    esac\n"
        f'    COMPREPLY=( $(compgen -W "${{words}}" -- ${{cur}}) )\n'
        f"}}\n"
        f"complete -F _loz loz\n"
    )
    print(completion)


if __name__ == "__main__":
    cli()
